/*-
 * Copyright (c) 2001, 2002 Allan Saddi <asaddi@philosophysw.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: yafic.c,v 1.33 2002/03/03 23:36:33 asaddi Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#ifdef HAVE_LIBGEN_H
#include <libgen.h>
#endif /* HAVE_LIBGEN_H */
#include <limits.h>
#ifdef DB_IN_DB1
#include <db1/db.h>
#else /* DB_IN_DB1 */
#include <db.h>
#endif /* DB_IN_DB1 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "yafic.h"
#include "statpack.h"
#include "sha1.h"

#include "version.h"

#ifndef lint
static const char rcsid[] =
	"$Id: yafic.c,v 1.33 2002/03/03 23:36:33 asaddi Exp $";
#endif /* !lint */

#ifndef HAVE_STRLCAT
size_t strlcat (char *dst, const char *src, size_t siz);
#endif /* !HAVE_STRLCAT */
#ifndef HAVE_DIRNAME
char *dirname (const char *path);
#endif /* !HAVE_DIRNAME */

int Verbosity;
int SimpleOutput;
char *prog;

#define HASH_BUFFER_SIZE 65536

static uint8_t *hashBuffer;
static DB *CheckDB, *UpdateDB;

void
yaficError (const char *what)
{
  fprintf (stderr, "%s: %s: %s\n", prog, what, strerror (errno));
  exit (1);
}

void *
mymalloc (size_t size)
{
  void *buf;

  if (!(buf = malloc (size)))
    yaficError ("malloc()");
  return buf;
}

char *
mystrdup (const char *str)
{
  char *newStr;

  if (!(newStr = strdup (str)))
    yaficError ("strdup()");
  return newStr;
}

uint32_t
myhash (const char *str)
{
  const uint8_t *s = str;
  uint32_t h = 0, g;

  while (*s) {
    h = (h << 4) + *(s++);
    if ((g = (h & 0xf0000000)))
      h ^= g >> 24;
    h &= ~g;
  }

  return h;
}

char *
mydirname (const char *path)
{
  char *tmp, *tmp2, *res;

  /* Just in case dirname() actually overwrites path. */
  tmp = mystrdup (path);
  if (!(tmp2 = dirname (tmp)))
    yaficError ("dirname");
  /* Just in case dirname() returns static storage. */
  res = mystrdup (tmp2);
  free (tmp);
  /* Caller is responsible for freeing result. */
  return res;
}

static void
hashDB (const char *dbName, uint8_t hash[SHA1_HASH_SIZE])
{
  SHA1Context sha;
  int f;
  ssize_t len;

  SHA1Init (&sha);

  if ((f = open (dbName, O_RDONLY)) != -1) {
    while ((len = read (f, hashBuffer, HASH_BUFFER_SIZE)) > 0)
      SHA1Update (&sha, hashBuffer, len);

    if (len < 0)
      yaficError (dbName);

    close (f);
  }
  else
    yaficError (dbName);

  SHA1Final (&sha, hash);
}

static int
hashFile (const char *path, struct stat *sb, uint8_t hash[SHA1_HASH_SIZE])
{
  SHA1Context sha;
  int f;
  ssize_t len;
  int error = 1;

  SHA1Init (&sha);

  if (S_ISLNK (sb->st_mode)) {
    if ((len = readlink (path, hashBuffer, HASH_BUFFER_SIZE)) != -1) {
      SHA1Update (&sha, hashBuffer, len);
      error = 0;
    }
  }
  else if (S_ISREG (sb->st_mode) || S_ISDIR (sb->st_mode)) {
    if ((f = open (path, O_RDONLY)) != -1) {
      while ((len = read (f, hashBuffer, HASH_BUFFER_SIZE)) > 0)
	SHA1Update (&sha, hashBuffer, len);
      close (f);
      error = len < 0;
    }
  }

  if (error)
    return 0;

  /* Compute final hash. */
  SHA1Final (&sha, hash);
  return 1;
}

const char *
toHexStr (uint8_t *data, int len)
{
  static const char hex[] = "0123456789abcdef";
  static char buf[128], *bufp;

  bufp = buf;
  while (len) {
    *(bufp++) = hex[*data >> 4];
    *(bufp++) = hex[*data & 0x0f];
    data++;
    len--;
  }
  *bufp = '\0';

  return buf;
}

static void
checkDeleted (const char *root, const char *checkDBName,
	      const char *updateDBName)
{
  DBT cKey, cData, uKey, uData;
  int res;
  int flags;
  char *entry;
  uint8_t *bufp;
  struct stat oldsb;

  char fullpath[PATH_MAX];

  memset (&cKey, 0, sizeof (cKey));
  memset (&cData, 0, sizeof (cData));
  memset (&uKey, 0, sizeof (uKey));
  memset (&uData, 0, sizeof (uData));

  if (!(CheckDB = dbopen (checkDBName, O_RDONLY, 0, DB_HASH, NULL)))
    yaficError (checkDBName);

  if (!(UpdateDB = dbopen (updateDBName, O_RDONLY, 0, DB_HASH, NULL)))
    yaficError (updateDBName);

  /* Iterate over each entry in the known database. */
  flags = R_FIRST;
  while (!CheckDB->seq (CheckDB, &cKey, &cData, flags)) {
    if (cData.size != DB_RECORD_SIZE) {
      fprintf (stderr, "%s: %s: %s\n", prog, checkDBName,
	       "bad database format");
      exit (1);
    }

    uKey.data = cKey.data;
    uKey.size = cKey.size;

    /* See if the entry still exists. */
    res = UpdateDB->get (UpdateDB, &uKey, &uData, 0);
    if (res == 1) { /* No longer exists. */
      entry = mymalloc (cKey.size + 1);
      strncpy (entry, cKey.data, cKey.size);
      entry[cKey.size] = '\0';

      snprintf (fullpath, sizeof (fullpath), "%s%s", root, entry);

      free (entry);

      /* Unpack the record. */
      UnpackStat (cData.data, &oldsb);
      bufp = &((uint8_t *) cData.data)[DB_RECORD_SIZE - SHA1_HASH_SIZE];

      switch (SimpleOutput) {
      case 0:
	printf ("removed: %s"
		" p(" ST_MODE_FMT ")"
		" i(" ST_INO_FMT ")"
		" n(" ST_NLINK_FMT ")"
		" u(" ST_UID_FMT ")"
		" g(" ST_GID_FMT ")"
		" s(" ST_SIZE_FMT ")"
		" a(" ST_ATIME_FMT ")"
		" m(" ST_MTIME_FMT ")"
		" c(" ST_CTIME_FMT ")"
		" h(%s)\n",
		fullpath,
		ST_MODE_CAST oldsb.st_mode,
		ST_INO_CAST oldsb.st_ino,
		ST_NLINK_CAST oldsb.st_nlink,
		ST_UID_CAST oldsb.st_uid,
		ST_GID_CAST oldsb.st_gid,
		ST_SIZE_CAST oldsb.st_size,
		ST_ATIME_CAST oldsb.st_atime,
		ST_MTIME_CAST oldsb.st_mtime,
		ST_CTIME_CAST oldsb.st_ctime,
		toHexStr (bufp, SHA1_HASH_SIZE));
	break;
      case 1:
	printf ("%s - REMOVED\n", fullpath);
	break;
      default:
	printf ("R %s\n", fullpath);
	break;
      }
    }
    else if (res)
      yaficError (updateDBName);

    flags = R_NEXT;
  }

  UpdateDB->close (UpdateDB);
  UpdateDB = NULL;

  CheckDB->close (CheckDB);
  CheckDB = NULL;
}

static void
checkFile (DB *db, const char *dbName, const char *fullpath,
	   const char *name, struct stat *sb, rflag_t flags)
{
  DBT key, data;
  int res;
  uint8_t *bufp;
  struct stat oldsb;
  int changed;
  int gothash;
  uint8_t hash[SHA1_HASH_SIZE];

  char buf[2 * PATH_MAX], buf2[256];

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  if (Verbosity >= 4)
    printf ("Checking %s...\n", fullpath);

  key.data = (char *) name;
  key.size = strlen (name);
  if (!(res = db->get (db, &key, &data, 0))) {
    if (data.size != DB_RECORD_SIZE) {
      fprintf (stderr, "%s: %s: %s\n", prog, dbName,
	       "bad database format");
      exit (1);
    }

    /* First unpack the record. */
    UnpackStat (data.data, &oldsb);
    bufp = &((uint8_t *) data.data)[DB_RECORD_SIZE - SHA1_HASH_SIZE];

    changed = 0;
    snprintf (buf, sizeof (buf), "changed: %s", fullpath);
    if ((flags & RFLAG_PERM) && sb->st_mode != oldsb.st_mode) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " p(" ST_MODE_FMT ":" ST_MODE_FMT ")",
		ST_MODE_CAST oldsb.st_mode, ST_MODE_CAST sb->st_mode);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_INODE) && sb->st_ino != oldsb.st_ino) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " i(" ST_INO_FMT ":" ST_INO_FMT ")",
		ST_INO_CAST oldsb.st_ino, ST_INO_CAST sb->st_ino);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_NLINK) && sb->st_nlink != oldsb.st_nlink) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " n(" ST_NLINK_FMT ":" ST_NLINK_FMT ")",
		ST_NLINK_CAST oldsb.st_nlink, ST_NLINK_CAST sb->st_nlink);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_UID) && sb->st_uid != oldsb.st_uid) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " u(" ST_UID_FMT ":" ST_UID_FMT ")",
		ST_UID_CAST oldsb.st_uid, ST_UID_CAST sb->st_uid);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_GID) && sb->st_gid != oldsb.st_gid) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " g(" ST_GID_FMT ":" ST_GID_FMT ")",
		ST_GID_CAST oldsb.st_gid, ST_GID_CAST sb->st_gid);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_SIZE) && sb->st_size != oldsb.st_size) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " s(" ST_SIZE_FMT ":" ST_SIZE_FMT ")",
		ST_SIZE_CAST oldsb.st_size, ST_SIZE_CAST sb->st_size);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_ATIME) && sb->st_atime != oldsb.st_atime) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " a(" ST_ATIME_FMT ":" ST_ATIME_FMT ")",
		ST_ATIME_CAST oldsb.st_atime, ST_ATIME_CAST sb->st_atime);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_MTIME) && sb->st_mtime != oldsb.st_mtime) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " m(" ST_MTIME_FMT ":" ST_MTIME_FMT ")",
		ST_MTIME_CAST oldsb.st_mtime, ST_MTIME_CAST sb->st_mtime);
      strlcat (buf, buf2, sizeof (buf));
    }
    if ((flags & RFLAG_CTIME) && sb->st_ctime != oldsb.st_ctime) {
      changed = 1;
      snprintf (buf2, sizeof (buf2), " c(" ST_CTIME_FMT ":" ST_CTIME_FMT ")",
		ST_CTIME_CAST oldsb.st_ctime, ST_CTIME_CAST sb->st_ctime);
      strlcat (buf, buf2, sizeof (buf));
    }
    if (flags & RFLAG_HASH) {
      if (!hashFile (fullpath, sb, hash))
	memset (hash, 0, sizeof (hash));

      if (memcmp (hash, bufp, SHA1_HASH_SIZE)) {
	changed = 1;
	snprintf (buf2, sizeof (buf2), " h(%s:",
		  toHexStr (bufp, SHA1_HASH_SIZE));
	strlcat (buf, buf2, sizeof (buf));
	snprintf (buf2, sizeof (buf2), "%s)",
		  toHexStr (hash, SHA1_HASH_SIZE));
	strlcat (buf, buf2, sizeof (buf));
      }
    }

    if (changed) {
      switch (SimpleOutput) {
      case 0:
	printf ("%s\n", buf);
	break;
      case 1:
	printf ("%s - CHANGED\n", fullpath);
	break;
      default:
	printf ("M %s\n", fullpath);
	break;
      }
    }
  }
  else if (res == 1) {
    gothash = (flags & RFLAG_HASH) && hashFile (fullpath, sb, hash);

    switch (SimpleOutput) {
    case 0:
      printf ("added: %s"
	      " p(" ST_MODE_FMT ")"
	      " i(" ST_INO_FMT ")"
	      " n(" ST_NLINK_FMT ")"
	      " u(" ST_UID_FMT ")"
	      " g(" ST_GID_FMT ")"
	      " s(" ST_SIZE_FMT ")"
	      " a(" ST_ATIME_FMT ")"
	      " m(" ST_MTIME_FMT ")"
	      " c(" ST_CTIME_FMT ")"
	      " h(%s)\n",
	      fullpath,
	      ST_MODE_CAST sb->st_mode,
	      ST_INO_CAST sb->st_ino,
	      ST_NLINK_CAST sb->st_nlink,
	      ST_UID_CAST sb->st_uid,
	      ST_GID_CAST sb->st_gid,
	      ST_SIZE_CAST sb->st_size,
	      ST_ATIME_CAST sb->st_atime,
	      ST_MTIME_CAST sb->st_mtime,
	      ST_CTIME_CAST sb->st_ctime,
	      gothash ? toHexStr (hash, SHA1_HASH_SIZE) : "-");
      break;
    case 1:
      printf ("%s - ADDED\n", fullpath);
      break;
    default:
      printf ("A %s\n", fullpath);
      break;
    }
  }
  else
    yaficError (dbName);
}

static void
updateFile (DB *db, const char *dbName, const char *fullpath,
	    const char *name, struct stat *sb, rflag_t flags)
{
  DBT key, data;
  uint8_t buf[128], *bufp;

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  if (Verbosity >= 4)
    printf ("Updating %s...\n", fullpath);

  PackStat (sb, buf);

  bufp = &buf[DB_RECORD_SIZE - SHA1_HASH_SIZE];
  if (!(flags & RFLAG_HASH) || !hashFile (fullpath, sb, bufp))
    memset (bufp, 0, SHA1_HASH_SIZE);
  bufp += 20;

  key.data = (char *) name;
  key.size = strlen (name);
  data.data = buf;
  data.size = DB_RECORD_SIZE;

  if (db->put (db, &key, &data, 0))
    yaficError (dbName);
}

static DB *scanQueue;

static void
dumpScanQueue (void)
{
  DBT key, data;
  int flags;
  int res;
  char *entry;

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  printf ("Scan queue:");
  flags = R_FIRST;
  while (!(res = scanQueue->seq (scanQueue, &key, &data, flags))) {
    entry = mymalloc (key.size + 1);
    strncpy (entry, key.data, key.size);
    entry[key.size] = '\0';

    printf (" %s", entry);

    free (entry);
    flags = R_NEXT;
  }
  printf ("\n");

  if (res == -1)
    yaficError ("db->seq()");
}

static void
scanFilesHelper (struct RuleEntry *re)
{
  DBT key, data;

  if (re->entryFlags & RFLAG_IGNORE)
    return;

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  key.data = re->path;
  key.size = strlen (re->path);
  data.data = "";
  data.size = 0;

  if (scanQueue->put (scanQueue, &key, &data, 0) == -1)
    yaficError ("db->put()");
}

static void
scanFiles (const char *root,
	   const char *checkDBName, const char *updateDBName,
	   int doCheck, int doUpdate)
{
  DBT key, data;
  char *entry;
  int res;

  struct RuleEntry *re;
  rflag_t flags;
  struct stat sb;

  int i;
  dev_t parentdev;
  DIR *dirp;
  struct dirent *dp;

  char fullpath[PATH_MAX];
  char descentry[PATH_MAX];
  char descpath[PATH_MAX];

  if (!doCheck && !doUpdate)
    return;

  if (doCheck &&
      !(CheckDB = dbopen (checkDBName, O_RDONLY, 0, DB_HASH, NULL)))
    yaficError (checkDBName);

  if (doUpdate &&
      !(UpdateDB = dbopen (updateDBName, O_CREAT|O_RDWR|O_TRUNC, 0600,
			   DB_HASH, NULL)))
    yaficError (updateDBName);

  /* Create a btree database for use as our priority queue. */
  if (!(scanQueue = dbopen (NULL, O_CREAT|O_RDWR|O_TRUNC, 0600, DB_BTREE,
			    NULL)))
    yaficError ("dbopen()");

  ApplyRuleSet (scanFilesHelper);

  memset (&key, 0, sizeof (key));
  memset (&data, 0, sizeof (data));

  /* The actual scanning loop. */
  for (;;) {
    if (Verbosity >= 3)
      dumpScanQueue ();

    if ((res = scanQueue->seq (scanQueue, &key, &data, R_FIRST)))
      break;
    entry = mymalloc (key.size + 1);
    strncpy (entry, key.data, key.size);
    entry[key.size] = '\0';
    if (scanQueue->del (scanQueue, &key, 0) == -1)
      yaficError ("db->del()");

    /* Process the entry itself. */
    if ((re = FindRuleEntry (entry)))
      flags = re->entryFlags;
    else {
      re = FindClosestRuleEntry (entry);
      flags = re->descFlags;
    }
    if (!(flags & RFLAG_IGNORE)) {
      snprintf (fullpath, sizeof (fullpath), "%s%s", root, entry);
      if (!lstat (fullpath, &sb)) {
	if (doCheck)
	  checkFile (CheckDB, checkDBName, fullpath, entry, &sb, flags);
	if (doUpdate)
	  updateFile (UpdateDB, updateDBName, fullpath, entry, &sb, flags);
      }
      else {
	free (entry);
	continue;
      }
    }
    else {
      free (entry);
      continue;
    }

    if (S_ISDIR (sb.st_mode)) {
      /* Now process its contents. */
      parentdev = sb.st_dev;
      re = FindClosestRuleEntry (entry);
      if (!(re->descFlags & RFLAG_IGNORE)) {
	if ((dirp = opendir (fullpath))) {
	  /* Kludge to keep paths clean. */
	  i = strlen (fullpath) - 1;
	  if (i >= 0 && fullpath[i] == '/')
	    fullpath[i] = '\0';
	  if (!strcmp (entry, "/"))
	    entry[0] = '\0';

	  while ((dp = readdir (dirp))) {
	    /* Skip "." and ".." */
	    if (!strcmp (dp->d_name, ".") || !strcmp (dp->d_name, ".."))
	      continue;

	    snprintf (descpath, sizeof (descpath), "%s/%s", fullpath,
		      dp->d_name);

	    if (!lstat (descpath, &sb)) {
	      snprintf (descentry, sizeof (descentry), "%s/%s", entry,
			dp->d_name);

	      if (!S_ISDIR (sb.st_mode)) {
		/* Optimize by copying above, skipping push/pop. */
		if ((re = FindRuleEntry (descentry)))
		  flags = re->entryFlags;
		else {
		  re = FindClosestRuleEntry (descentry);
		  flags = re->descFlags;
		}
		if (!(flags & RFLAG_IGNORE)) {
		  if (doCheck)
		    checkFile (CheckDB, checkDBName, descpath, descentry,
			       &sb, flags);
		  if (doUpdate)
		    updateFile (UpdateDB, updateDBName, descpath, descentry,
				&sb, flags);
		}
	      }
	      else {
		if (sb.st_dev == parentdev) {
		  key.data = descentry;
		  key.size = strlen (descentry);
		  data.data = "";
		  data.size = 0;
		  if (scanQueue->put (scanQueue, &key, &data, R_NOOVERWRITE)
		      == -1)
		    yaficError ("db->put()");
		}
	      }
	    }
	  }

	  closedir (dirp);
	}
      }
    }

    free (entry);
  }

  scanQueue->close (scanQueue);
  scanQueue = NULL;

  if (UpdateDB) {
    res = UpdateDB->close (UpdateDB);
    UpdateDB = NULL;
    if (res)
      yaficError (updateDBName);
  }

  if (CheckDB) {
    CheckDB->close (CheckDB);
    CheckDB = NULL;
  }
}

static void
cleanUp (void)
{
  if (scanQueue)
    scanQueue->close (scanQueue);
  if (UpdateDB)
    UpdateDB->close (UpdateDB);
  if (CheckDB)
    CheckDB->close (CheckDB);
  CleanRuleSet ();
  free (hashBuffer);
}

static char *
filler (char *template)
{
  int len;
  static char spaces[128];

  len = strlen (template);
  if (len > (sizeof (spaces) - 1))
    len = sizeof (spaces) - 1;
  memset (spaces, ' ', len);
  spaces[len] = '\0';

  return spaces;
}

static void
help (void)
{
  fprintf (stderr,
	   "\nOptions:\n"
	   "\t-V\t\tDisplay version information.\n"
	   "\t-h\t\tDisplay this summary.\n"
	   "\t-v\t\tIncrease verbosity.\n"
	   "\t-s\t\tShow simple output.\n"
	   "\t-C <config>\tSpecify configuration file. (Default: %s)\n"
	   "\t-r <root>\tSpecify root. (Default: %s)\n"
	   "\t-c <database>\tCompare with known database. Specify - to use\n"
	   "\t\t\tdefault database. (Default: %s)\n"
	   "\t-u <database>\tPerform update. Specify - to use default\n"
	   "\t\t\tdatabase. (Default: %s)\n"
	   "\t-H\t\tDisplay SHA-1 hash of updated database.\n"
	   "\t-l <database>\tList entries in database.\n"
	   "\t-t <type>\tList only files of a specific type. <type> may be\n"
	   "\t\t\td, f, l, or s or any combination thereof.\n"
	   "\t\t\t(Directory, file, symlink, special, respectively.)\n"
	   "\n"
	   "  The -v option may be specified multiple times. More than twice\n"
	   "  will only slow things down! If used with -l, the actual\n"
	   "  database contents will also be displayed.\n"
	   "\n"
	   "  The -s option may be specified twice, giving simpler output.\n"
	   "\n"
	   "  To detect removed files, the -c and -u options must be used\n"
	   "  together.\n"
	   , DEFAULT_CONF, /* DEFAULT_ROOT */ "/",
	   DEFAULT_CHECK_DB, DEFAULT_UPDATE_DB);
}

static void
usage (void)
{
  fprintf (stderr,
	   "Usage: %s [-Vhvs] [-C <config>] [-r <root>] [-c <known-database>]\n"
	   "       %s [-u <new-database> [-H]]\n"
	   "       %s [-Vhv] [-r <root>] [-t <type>] -l <database>\n",
	   prog, filler (prog), prog);
}

int
main (int argc, char *argv[])
{
  int ch;
  int doCheck, doUpdate, doView, doHash;
  char *conf, *root, *checkDB, *updateDB, *viewDB;
  char *typeStr;
  int i;

  uint8_t hash[SHA1_HASH_SIZE];

  doCheck = doUpdate = doView = doHash = 0;
  conf = DEFAULT_CONF;
  root = DEFAULT_ROOT;
  checkDB = DEFAULT_CHECK_DB;
  updateDB = DEFAULT_UPDATE_DB;
  viewDB = NULL;
  typeStr = NULL;

  prog = argv[0];
  while ((ch = getopt (argc, argv, "C:HVc:hl:r:st:u:v")) != -1) {
    switch (ch) {
    case 'C':
      conf = optarg;
      break;
    case 'H':
      doHash++;
      break;
    case 'V':
      fprintf (stderr, "yafic version " VERSION_STRING
	       " (" VERSION_DATE ")\n");
      exit (1);
    case 'c':
      doCheck = 1;
      if (strcmp (optarg, "-"))
	checkDB = optarg;
      break;
    case 'h':
      usage ();
      help ();
      exit (2);
    case 'l':
      doView = 1;
      viewDB = optarg;
      break;
    case 'r':
      if (!strcmp (optarg, "/"))
	root = "";
      else {
	root = optarg;
	/* Wipe out trailing slash, if any. */
	i = strlen (root) - 1;
	if (i > 1 && root[i] == '/')
	  root[i] = '\0';
      }
      break;
    case 's':
      SimpleOutput++;
      break;
    case 't':
      typeStr = optarg;
      break;
    case 'u':
      doUpdate = 1;
      if (strcmp (optarg, "-"))
	updateDB = optarg;
      break;
    case 'v':
      Verbosity++;
      break;
    case '?':
    default:
      usage ();
      exit (2);
    }
  }
  argc -= optind;
  argv += optind;

  if (argc || (doView && (doCheck || doUpdate))) {
    usage ();
    exit (2);
  }

  if (doView) {
    ViewDB (root, viewDB, typeStr);
    exit (0);
  }

  if (Verbosity) {
    printf ("        config: %s\n", conf);
    printf ("          root: %s\n", !strlen (root) ? "/" : root);
    printf ("known-database: %s\n", checkDB);
    printf ("  new-database: %s\n", updateDB);
    printf ("     operation: %s\n",
	    doCheck && doUpdate ? "check and update" :
	    (doCheck ? "check" :
	     (doUpdate ? "update" : "none")));
  }

  atexit (cleanUp);

  hashBuffer = mymalloc (HASH_BUFFER_SIZE);
  InitRuleSet ();

  if (!ParseRuleSet (conf))
    exit (1);

  if (Verbosity >= 2) {
    printf ("\nParsed rule set:\n");
    ApplyRuleSet (DumpRuleEntry);

    if (Verbosity >= 3)
      printf ("\n");
  }

  /* Do the actual check/update. */
  scanFiles (root, checkDB, updateDB, doCheck, doUpdate);
  /* Check for deleted files, if applicable. */
  if (doCheck && doUpdate)
    checkDeleted (root, checkDB, updateDB);

  if (doUpdate && doHash) {
    hashDB (updateDB, hash);
    printf ("%s %s\n", toHexStr (hash, SHA1_HASH_SIZE), updateDB);
  }

  /* If neither checking nor updating and not verbose, say something to
     reassure the user. */
  if (!doCheck && !doUpdate && !Verbosity)
    printf ("nothing done; use -c, -u, or -l options\n");

  exit (0);
}
